/***************************************************
  STM32 Support added by Jaret Burkett at OSHlab.com

  This is our library for the Adafruit  ILI9488 Breakout and Shield
  ----> http://www.adafruit.com/products/1651

  Check out the links above for our tutorials and wiring diagrams
  These displays use SPI to communicate, 4 or 5 pins are required to
  interface (RST is optional)
  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  Written by Limor Fried/Ladyada for Adafruit Industries.
  MIT license, all text above must be included in any redistribution
 ****************************************************/

#ifndef _ILI9488H_
#define _ILI9488H_

#include "Arduino.h"
#include "Adafruit_GFX.h"
#include <Adafruit_SPITFT.h>
#include <SPI.h>

#define ILI9488_TFTWIDTH  320	///< ILI9488 max TFT width
#define ILI9488_TFTHEIGHT 480	///< ILI9488 max TFT height

#define ILI9488_NOP 0x00     ///< No-op register
#define ILI9488_SWRESET 0x01 ///< Software reset register
#define ILI9488_RDDID 0x04   ///< Read display identification information
#define ILI9488_RDDST 0x09   ///< Read Display Status

#define ILI9488_SLPIN 	0x10  ///< Enter Sleep Mode
#define ILI9488_SLPOUT 	0x11 ///< Sleep Out
#define ILI9488_PTLON 	0x12  ///< Partial Mode ON
#define ILI9488_NORON 	0x13  ///< Normal Display Mode ON

#define ILI9488_RDMODE 	0x0A     ///< Read Display Power Mode
#define ILI9488_RDMADCTL 0x0B   ///< Read Display MADCTL
#define ILI9488_RDPIXFMT 0x0C   ///< Read Display Pixel Format
#define ILI9488_RDIMGFMT 0x0D   ///< Read Display Image Format
#define ILI9488_RDSELFDIAG 0x0F ///< Read Display Self-Diagnostic Result

#define ILI9488_INVOFF 	0x20   ///< Display Inversion OFF
#define ILI9488_INVON 	0x21    ///< Display Inversion ON
#define ILI9488_GAMMASET 0x26 ///< Gamma Set
#define ILI9488_DISPOFF 0x28  ///< Display OFF
#define ILI9488_DISPON 	0x29   ///< Display ON

#define ILI9488_CASET 	0x2A	///< Column Address Set
#define ILI9488_PASET 	0x2B	///< Page Address Set
#define ILI9488_RAMWR 	0x2C	///< Memory Write
#define ILI9488_RAMRD 	0x2E	///< Memory Read

#define ILI9488_PTLAR 	0x30	///< Partial Area
#define ILI9488_VSCRDEF 0x33	///< Vertical Scrolling Definition
#define ILI9488_MADCTL 	0x36	///< Memory Access Control
#define ILI9488_VSCRSADD 0x37	///< Vertical Scrolling Start Address
#define ILI9488_PIXFMT 	0x3A	///< COLMOD: Pixel Format Set

#define ILI9488_FRMCTR1 0xB1	///< Frame Rate Control (In Normal Mode/Full Colors)
#define ILI9488_FRMCTR2 0xB2	///< Frame Rate Control (In Idle Mode/8 colors)
#define ILI9488_FRMCTR3 0xB3	///< Frame Rate control (In Partial Mode/Full Colors)
#define ILI9488_INVCTR	0xB4	///< Display Inversion Control
#define ILI9488_DFUNCTR 0xB6	///< Display Function Control

#define ILI9488_PWCTR1	0xC0	///< Power Control 1
#define ILI9488_PWCTR2	0xC1 	///< Power Control 2
#define ILI9488_PWCTR3	0xC2 	///< Power Control 3
#define ILI9488_PWCTR4	0xC3	///< Power Control 4
#define ILI9488_PWCTR5	0xC4	///< Power Control 5
#define ILI9488_VMCTR1	0xC5	///< VCOM Control 1
#define ILI9488_VMCTR2	0xC7	///< VCOM Control 2

#define ILI9488_RDID1   0xDA	///< Read ID 1
#define ILI9488_RDID2   0xDB	///< Read ID 2
#define ILI9488_RDID3   0xDC	///< Read ID 3
#define ILI9488_RDID4   0xDD	///< Read ID 4

#define ILI9488_GMCTRP1 0xE0	///< Positive Gamma Correction
#define ILI9488_GMCTRN1 0xE1	///< Negative Gamma Correction
//#define ILI9488_PWCTR6  0xFC
#define ILI_SPI_CHIP_SEL 3

// Color definitions
#define ILI9488_BLACK       0x0000      /*   0,   0,   0 */
#define ILI9488_NAVY        0x000F      /*   0,   0, 128 */
#define ILI9488_DARKGREEN   0x03E0      /*   0, 128,   0 */
#define ILI9488_DARKCYAN    0x03EF      /*   0, 128, 128 */
#define ILI9488_MAROON      0x7800      /* 128,   0,   0 */
#define ILI9488_PURPLE      0x780F      /* 128,   0, 128 */
#define ILI9488_OLIVE       0x7BE0      /* 128, 128,   0 */
#define ILI9488_LIGHTGREY   0xC618      /* 192, 192, 192 */
#define ILI9488_DARKGREY    0x7BEF      /* 128, 128, 128 */
#define ILI9488_BLUE        0x001F      /*   0,   0, 255 */
#define ILI9488_GREEN       0x07E0      /*   0, 255,   0 */
#define ILI9488_CYAN        0x07FF      /*   0, 255, 255 */
#define ILI9488_RED         0xF800      /* 255,   0,   0 */
#define ILI9488_MAGENTA     0xF81F      /* 255,   0, 255 */
#define ILI9488_YELLOW      0xFFE0      /* 255, 255,   0 */
#define ILI9488_WHITE       0xFFFF      /* 255, 255, 255 */
#define ILI9488_ORANGE      0xFD20      /* 255, 165,   0 */
#define ILI9488_GREENYELLOW 0xAFE5      /* 173, 255,  47 */
#define ILI9488_PINK        0xF81F

class ILI9488 : public Adafruit_GFX
{
	public:
		ILI9488(int8_t _CS, int8_t _DC, int8_t _MOSI, int8_t _SCLK, int8_t _RST, int8_t _MISO);
		ILI9488(int8_t _CS, int8_t _DC, int8_t _RST = -1);
		
		void begin(void);
		void setAddrWindow(uint16_t x0, uint16_t y0, uint16_t x1, uint16_t y1);
		void setScrollArea(uint16_t topFixedArea, uint16_t bottomFixedArea);
		void scroll(uint16_t pixels);
		void pushColor(uint16_t color);
		void pushColors(uint16_t *data, uint8_t len, boolean first);
		void drawImage(const uint8_t* img, uint16_t x, uint16_t y, uint16_t w, uint16_t h);
		void fillScreen0(uint16_t color);
		void fillScreen(uint16_t color);
		void drawPixel(int16_t x, int16_t y, uint16_t color);
		void drawFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color);
		void drawFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color);
		void fillRect0(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color);
		void fillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color);
		void setRotation(uint8_t r);
		void invertDisplay(boolean i);
		void pushRGB(uint8_t r, uint8_t g, uint8_t b);
		
		// New.. yet to be implemented
		//void begin(uint32_t freq = 0);
		//void setRotation(uint8_t r);
		//void invertDisplay(bool i);
		//void scrollTo(uint16_t y);
		//void setScrollMargins(uint16_t top, uint16_t bottom);
		
		uint16_t color565(uint8_t r, uint8_t g, uint8_t b);

		/* These are not for current use, 8-bit protocol only! */
		uint8_t  readdata(void);
		uint8_t  readcommand8(uint8_t reg, uint8_t index = 0);

		/*
		uint16_t readcommand16(uint8_t);
		uint32_t readcommand32(uint8_t);
		void     dummyclock(void);
		*/

		void spiwrite(uint8_t);
		void writecommand(uint8_t c);
		void write16BitColor(uint16_t color);
		void writedata(uint8_t d);
		void commandList(uint8_t *addr);

		uint8_t  spiread(void);

	private:
		uint8_t  tabcolor;

		boolean  hwSPI;
		
		__attribute__((always_inline))
		void digitalWriteDirect(int pin, boolean val)
		{
			if(val)
				g_APinDescription[pin].pPort -> PIO_SODR = g_APinDescription[pin].ulPin;
			else
				g_APinDescription[pin].pPort -> PIO_CODR = g_APinDescription[pin].ulPin;
		}
		
		__attribute__((always_inline))
		int digitalReadDirect(int pin)
		{
			return !!(g_APinDescription[pin].pPort -> PIO_PDSR & g_APinDescription[pin].ulPin);
		}
	
		/** SPI send a byte */
		__attribute__((always_inline))
		void dmaSend(uint8_t b)
		{
			dmaSpiTransfer(b);
		}
	
		__attribute__((always_inline))
		uint8_t dmaSpiTransfer(uint8_t b)
		{
			Spi* pSpi = SPI0;
			
			pSpi->SPI_TDR = b;
			while ((pSpi->SPI_SR & SPI_SR_RDRF) == 0) {}
			b = pSpi->SPI_RDR;
			return b;
		}
		
		//  initialize SPI controller
		void dmaInit(uint8_t sckDivisor)
		{
			uint8_t scbr = sckDivisor;
			Spi* pSpi = SPI0;
			//  disable SPI
			pSpi->SPI_CR = SPI_CR_SPIDIS;
			// reset SPI
			pSpi->SPI_CR = SPI_CR_SWRST;
			// no mode fault detection, set master mode
			pSpi->SPI_MR = SPI_PCS(ILI_SPI_CHIP_SEL) | SPI_MR_MODFDIS | SPI_MR_MSTR;
			// mode 0, 8-bit,
			pSpi->SPI_CSR[ILI_SPI_CHIP_SEL] = SPI_CSR_SCBR(scbr) | SPI_CSR_NCPHA | SPI_CSR_BITS_8_BIT;
			// enable SPI
			pSpi->SPI_CR |= SPI_CR_SPIEN;
		}

		#if defined (__AVR__) || defined(TEENSYDUINO)
		  uint8_t mySPCR;
		  volatile uint8_t *mosiport, *clkport, *dcport, *rsport, *csport;
		  int8_t  _cs, _dc, _rst, _mosi, _miso, _sclk;
		  uint8_t  mosipinmask, clkpinmask, cspinmask, dcpinmask;
		////This def is for the Arduino.ORG M0!!!
		//#elif defined(ARDUINO_SAM_ZERO)
		//    volatile PORT_OUT_Type *mosiport, *clkport, *dcport, *rsport, *csport;
		//    int32_t  _cs, _dc, _rst, _mosi, _miso, _sclk;
		//    PORT_OUT_Type  mosipinmask, clkpinmask, cspinmask, dcpinmask;
		#elif defined (__STM32F1__) || defined (_VARIANT_ARDUINO_STM32_) || defined (STM32F100xE) || defined (STM32F101xE) || defined (STM32F101xG) || defined (STM32F103xE) || defined (STM32F103xG) || defined (STM32F105xC) || defined (STM32F107xC)
			uint8_t mySPCR;
			volatile uint32_t *mosiport, *clkport, *dcport, *rsport, *csport;
			int32_t  _cs, _dc, _rst, _mosi, _miso, _sclk;
			uint32_t  mosipinmask, clkpinmask, cspinmask, dcpinmask;
		#elif defined (__arm__)
			volatile RwReg *mosiport, *clkport, *dcport, *rsport, *csport;
			int32_t  _cs, _dc, _rst, _mosi, _miso, _sclk;
			uint32_t  mosipinmask, clkpinmask, cspinmask, dcpinmask;
		#elif defined (ESP8266)
			int32_t  _cs, _dc, _rst, _mosi, _miso, _sclk;
		#else
			int8_t  _cs, _dc, _rst, _mosi, _miso, _sclk;
		#endif		
};
#endif
